/*
 * 
 * 
 */

package autofeature.intermediate;

import autofeature.ParserBug;
import autofeature.intermediate.State;
import java.util.ArrayList;
import java.util.List;
import org.antlr.runtime.CommonToken;
import org.antlr.runtime.Token;
import org.antlr.runtime.tree.CommonTree;
import static autofeature.intermediate.TreeUtil.*;


/**
 *
 * @author Hendrik Speidel - speidel@fmi.uni-passau.de
 */
public class Automaton {

    private String introduction;

    private List<Event> globalEvents;

    private List<State> states;

    private List<Shadow> shadows;

    private final AST ast;

    public Automaton(AST ast) {
        this.ast = ast;
    }

    public String getComment() {

        int l = ast.root.getTokenStartIndex() - 1;
        //System.out.println(l);

        while (l >= 0) {
            CommonToken t = (CommonToken) ast.input.get(l);
            if (t == null || t.getChannel() != CommonToken.HIDDEN_CHANNEL) {
                break;
            } else {
                l--;
            }
        }

        if (l+1 >= ast.root.getTokenStartIndex()) {
            return "";
        } else {
            return ast.input.toString(l + 1, ast.root.getTokenStartIndex() - 1);
        }

    }

    public String getName() {
        return ast.root.getChild(0).getText();
    }

    public List<Event> getGlobalEvents() {

        if (globalEvents == null) {
            globalEvents = new ArrayList<Event>();
            CommonTree globalEcas = (CommonTree) ast.root.getChild(2);
            if (!globalEcas.getText().equals("GLOBAL_ECA")) {
                throw new ParserBug("3rd Child of automaton must be GLOBAL_ECA!");
            }
            
            for (CommonTree child : childs(globalEcas)) {
                globalEvents.add(new Event(ast.asAst(child)));
            }
        }
        return globalEvents;
    }

    public List<State> getStates() {
        if (states == null) {
            states = new ArrayList<State>();
            for (CommonTree state: childs(ast.root, 3)) {
                states.add(new State(this, ast.asAst(state)));
            }
        }
        return states;
    }

    public String getIntroduction() {
        if (introduction == null) {
            CommonTree intro = (CommonTree) ast.root.getChild(1);
            introduction = ast.blocks.get(intro);
            if (introduction == null) {
                return "";
                //throw new ParserBug("INTRODUCTION but no block attached!");
            }
        }
        return introduction;
    }

    public List<Shadow> getShadows() {
        if (shadows == null) {
            shadows = new ArrayList<Shadow>();

            for (CommonTree t : childs((CommonTree) ast.root.getChild(1))) {
                if (!t.getText().equals("shadow")) {
                    break;
                }
                shadows.add(new Shadow(ast.asAst(t)));
            }

        }
        return shadows;
    }
}
