package autofeature.intermediate;

import autofeature.ParserBug;
import java.util.ArrayList;
import java.util.List;
import org.antlr.runtime.tree.CommonTree;
import static autofeature.intermediate.TreeUtil.*;

public class Event {

    private Boolean before;
    private String condition;
    private String action;
    private String returnType;
    private List<String> parameterTypes;
    private List<String> parameterNames;
    private final AST ast;

    public Event(AST ast) {
        this.ast = ast;
    }

    public boolean isBefore() {
        if (before == null) {
            before = ast.root.getText().equals("before");
        }
        return before;
    }

    public String getReturnType() {
        if (returnType == null) {
            returnType = extractParameterType(ast.root);
        }
        return returnType;
    }

    public List<String> getParameterTypes() {
        if (parameterTypes == null) {
            parameterTypes = new ArrayList<String>();
            CommonTree paramlist = (CommonTree) getDeclarator().getChild(getDeclarator().getChildCount() - 1);
            for (CommonTree param : childs(paramlist)) {
                String paramType = extractParameterType(param);
                parameterTypes.add(paramType);
            }
        }
        return parameterTypes;
    }

    public List<String> getParameterNames() {
        if (parameterNames == null) {
            parameterNames = new ArrayList<String>();
            CommonTree paramlist = (CommonTree) getDeclarator().getChild(getDeclarator().getChildCount() - 1);
            for (CommonTree param : childs(paramlist)) {
                String paramName = extractParameterName(param);
                parameterNames.add(paramName);
            }
        }
        return parameterNames;
    }

    private String extractParameterName(CommonTree declaration) {
        String first = declaration.getChild(1).getChild(0).getText();
        if (first.equals("*")) {
            return declaration.getChild(1).getChild(1).getText();
        } else {
            return first;
        }
    }

    private String extractParameterType(CommonTree declaration) {
        String type = joinChilds((CommonTree) declaration.getChild(0));
        if (declaration.getChild(1).getChild(0).getText().equals("*")) {
            type += " *";
        }
        return type;
    }

    private CommonTree getDeclarator() {
        return (CommonTree) ast.root.getChild(1);
    }

    public String getFunctionName() {
        return extractParameterName(ast.root);
    }

    public String getCondition() {
        if (condition == null) {
            CommonTree cond = (CommonTree) ast.root.getChild(2);
            if (!cond.getText().equals("EVENT_CONDITION")) {
                cond = (CommonTree) ast.root.getChild(3);
                //throw new ParserBug("event.getChild(3) is not a EVENT_CONDITION!");
            }
            condition = ast.blocks.get(cond);
            
        }
        return condition;
    }

    public String getAction() {
        if (action == null) {
            CommonTree ac = (CommonTree) ast.root.getChild(3);
            if (!ac.getText().equals("COMPOUND_STATEMENT")) {
                ac = (CommonTree) ast.root.getChild(4);
                //throw new ParserBug("event.getChild(3) is not a COMPOUND_STATEMENT!");
            }
            action = ast.blocks.get(ac).trim();
            if (action.startsWith("{")) {
                action = action.substring(1, action.length() - 1);
            }
            if (action == null) {
                throw new ParserBug("event COMPOUND_STATEMENT has no block attached!");
            }
        }
        return action;
    }

    public String getReturnValueName() {
        if (ast.root.getChild(2).getText().equals("RETVAL")) {
            return ast.root.getChild(2).getChild(0).getText();
        }
        return null;


    }

}
